/*  einsc -- a compiler for eins
 *    Code generator
 *
 *    Michel Schinz
 *    $Id: Generator.java,v 1.1 2005/03/15 14:46:53 dragos Exp $
 */

package einsc;

import java.util.BitSet;

import einsc.Tree.*;

public class Generator {
    private final Analyzer analyzer;

    private final Gen genVisitor;
    private final GenLoad genLoadVisitor;
    private final GenCond genCondVisitor;
    public final Code code;

    public Generator(Analyzer analyzer) {
        this.analyzer = analyzer;
        this.genVisitor = new Gen();
        this.genLoadVisitor = new GenLoad();
        this.genCondVisitor = new GenCond();
        this.code = getCode();
    }

    public Code getCode() { return new Code(); }
    
    public void gen(Tree tree) {
        genVisitor.generate(tree);
    }

    public void gen(Tree[] trees) {
        for (int i = 0; i < trees.length; ++i)
            gen(trees[i]);
    }

    public void genLoad(Tree tree, int targetReg) {
        assert code.usedRegisters().get(targetReg)
            : "invariant violated: target register R" + targetReg
            + " not allocated";
        genLoadVisitor.generate(tree, targetReg);
    }

    public void genCond(Tree tree, Code.Label targetLabel, boolean when) {
        genCondVisitor.generate(tree, targetLabel, when);
    }

    /**
     * Visitor generating code for statements, in a general sense,
     * that is all constructs which do not have a value.
     */
    private class Gen extends DefaultVisitor implements Visitor, RISC {

        public void generate(Tree tree) {
            tree.apply(this);
        }

        public void caseProgram(Program tree) {
            // initialise stack pointer
            code.emit(SYSCALL, SP, 0, SYS_GET_TOTAL_MEM_SIZE);
            // initialise garbage collector
            code.emit(DIVI, 1, SP, 4);
            code.emit(DIVI, 2, 1, 2);
            code.emit(ADDI, 3, 0, SP);
            code.emit(LSHI, 3, 3, 27);
            code.emit(ADD, 2, 2, 3);
            code.emit(SYSCALL, 1, 2, SYS_GC_INIT);

            // jump to main expression
            Code.Label startLabel = code.getLabel();
            code.emit(BEQ, ZERO, startLabel);

            // generate predefined functions
            for (Symbol sym = analyzer.globalScope.first(); sym != null; sym = sym.next) {
                sym.offset = code.pc();
                if (sym.name.equals("printInt")) {
                    code.emit(POP, 1, SP, WORD_SIZE);
                    code.emit(SYSCALL, 1, ZERO, SYS_IO_WR_INT);
                    code.emit(ADD, RES, ZERO, ZERO);
                } else if (sym.name.equals("printChar")) {
                    code.emit(POP, 1, SP, WORD_SIZE);
                    code.emit(SYSCALL, 1, ZERO, SYS_IO_WR_CHR);
                    code.emit(ADD, RES, ZERO, ZERO);
                } else if (sym.name.equals("readInt")) {
                    code.emit(SYSCALL, RES, ZERO, SYS_IO_RD_INT);
                } else if (sym.name.equals("readChar")) {
                    code.emit(SYSCALL, RES, ZERO, SYS_IO_RD_CHR);
                };
                code.emit(RET, LNK);
            }

            // generate predefined classes
            Symbol listClass = analyzer.globalScope.lookup("List");
            Symbol[] fields = ((Type.ClassType) listClass.type).fields.toArray();
            for (int i = 0; i < fields.length; i++)
                fields[i].offset = i * WORD_SIZE;

            // generate class and function definitions
            gen(tree.decls);

            // generate main expression
            code.anchorLabel(startLabel);
            int tmpReg = code.getRegister();
            genLoad(tree.main, tmpReg);
            code.freeRegister(tmpReg);

            // exit
            code.emit(RET, ZERO);
        }

        public void caseClassDef(ClassDef tree) {
            for(int i = 0; i < tree.fields.length; i++)
                ((Formal) tree.fields[i]).sym.offset = i * WORD_SIZE;
        }

        public void caseFunDef(FunDef tree) {
            tree.sym.offset = code.pc();
            gen(tree.params);
            code.incFrameSize(WORD_SIZE);
            code.emit(PSH, LNK, SP, WORD_SIZE);
            int resReg = code.getRegister(RES);
            genLoad(tree.body, resReg);
            int popSize = WORD_SIZE * (1 + tree.params.length);
            code.emit(POP, LNK, SP, popSize);
            code.decFrameSize(popSize);
            assert code.getFrameSize() == 0
                : "non-empty frame (size " + code.getFrameSize() + ")";
            code.emit(RET, LNK);
            code.freeRegister(resReg);
        }

        public void caseFormal(Formal tree) {
            code.incFrameSize(WORD_SIZE);
            tree.sym.offset = code.getFrameSize();
        }

        public void caseWhile(While tree) {
            Code.Label startLabel = code.getLabel();
            code.anchorLabel(startLabel);
            Code.Label afterLabel = code.getLabel();
            genCond(tree.cond, afterLabel, false);
            int beginFrameSize = code.getFrameSize();
            gen(tree.stats);
            int deltaFrameSize = code.getFrameSize() - beginFrameSize;
            if (deltaFrameSize > 0) {
                code.emit(ADDI, SP, SP, deltaFrameSize);
                code.decFrameSize(deltaFrameSize);
            }
            code.emit(BEQ, ZERO, startLabel);
            code.anchorLabel(afterLabel);
        }

        public void caseVarDef(VarDef tree) {
            int tmpReg = code.getRegister();
            genLoad(tree.init, tmpReg);
            gen(tree.variable);
            code.emit(PSH, tmpReg, SP, WORD_SIZE);
            code.freeRegister(tmpReg);
        }

        public void caseAssign(Assign tree) {
            int tmpReg = code.getRegister();
            genLoad(tree.expr, tmpReg);
            int realOffset = code.getFrameSize() - tree.sym.offset;
            code.emit(STW, tmpReg, SP, realOffset);
            code.freeRegister(tmpReg);
        }

        public void caseDo(Do tree) {
            int tmpReg = code.getRegister();
            genLoad(tree.expr, tmpReg);
            code.freeRegister(tmpReg);
        }

        public void caseDefault(Tree tree) {
            throw new Error("unexpected case");
        }
    }

    /**
     * Visitor generating code for expressions, that is constructs
     * which produce a value.
     */
    private class GenLoad extends DefaultVisitor implements Visitor, RISC {
        private int targetReg;

        public void generate(Tree tree, int targetReg) {
            int bkpTargetReg = this.targetReg;
            this.targetReg = targetReg;
            tree.apply(this);
            this.targetReg = bkpTargetReg;
        }

        public void caseIf(If tree) {
            Code.Label elseLabel = code.getLabel();
            genCond(tree.cond, elseLabel, false);
            genLoad(tree.thenp, targetReg);
            Code.Label afterLabel = code.getLabel();
            code.emit(BEQ, ZERO, afterLabel);
            code.anchorLabel(elseLabel);
            genLoad(tree.elsep, targetReg);
            code.anchorLabel(afterLabel);
        }

        public void caseBinop(Binop tree) {
            switch (tree.op) {
            case Tree.ADD: case Tree.SUB:
            case Tree.MUL: case Tree.DIV: case Tree.MOD: {
                genLoad(tree.left, targetReg);
                int rightReg = code.getRegister();
                genLoad(tree.right, rightReg);
                
                if (tree.inversed)
                    code.emit(opcode(tree.op), targetReg, rightReg, targetReg);
                else
                	code.emit(opcode(tree.op), targetReg, targetReg, rightReg);
                
                code.freeRegister(rightReg);
            } break;
            default:
                caseDefault(tree);
            }
        }

        public void caseIdent(Ident tree) {
            if (tree.sym.kind == Symbol.VAR) {
                int realOffset = code.getFrameSize() - tree.sym.offset;
                code.emit(LDW, targetReg, SP, realOffset);
            } else
                emitLoadConstant(targetReg, tree.sym.offset);
        }


        public void caseNullLit(NullLit tree) {
            emitLoadConstant(targetReg, 0);
        }

        public void caseIntLit(IntLit tree) {
            emitLoadConstant(targetReg, tree.value);
        }

        public void caseApply(Apply tree) {
            int funAddr = -1;
            int funReg = -1;
            if ((tree.fun instanceof Ident)
                && ((Ident)tree.fun).sym.kind == Symbol.FUNC)
                funAddr = ((Ident)tree.fun).sym.offset;

            // save registers
            code.freeRegister(targetReg);
            BitSet usedRegs = code.usedRegisters();
            for (int r = usedRegs.nextSetBit(0);
                 r >= 0;
                 r = usedRegs.nextSetBit(r+1)) {
                    code.incFrameSize(WORD_SIZE);
                    code.emit(PSH, r, SP, WORD_SIZE);
                    code.freeRegister(r);
            }

            if (funAddr == -1) {
                funReg = code.getRegister();
                genLoad(tree.fun, funReg);
            }

            // load arguments
            int tmpReg = code.getRegister();
            for (int i = 0; i < tree.args.length; ++i) {
                genLoad(tree.args[i], tmpReg);
                code.incFrameSize(WORD_SIZE);
                code.emit(PSH, tmpReg, SP, WORD_SIZE);
            }
            code.freeRegister(tmpReg);

            if (funAddr != -1)
                code.emit(JSR, funAddr / WORD_SIZE);
            else {
                // load the return address in the Link register
                emitLoadConstant(LNK, code.pc() + WORD_SIZE * 2);
                code.emit(RET, funReg);
                code.freeRegister(funReg);
            }

            code.decFrameSize(tree.args.length * WORD_SIZE);
            code.getRegister(targetReg);
            if (targetReg != RES)
                code.emit(ADDI, targetReg, RES, 0);

            // restore registers
            for (int r = usedRegs.nextSetBit(0);
                 r >= 0;
                 r = usedRegs.nextSetBit(r+1)) {
                    code.getRegister(r);
                    code.emit(POP, r, SP, WORD_SIZE);
                    code.decFrameSize(WORD_SIZE);
            }
        }

        public void caseBlock(Block tree) {
            int beginFrameSize = code.getFrameSize();
            gen(tree.stats);
            genLoad(tree.main, targetReg);
            int deltaFrameSize = code.getFrameSize() - beginFrameSize;
            if (deltaFrameSize > 0) {
                code.emit(ADDI, SP, SP, deltaFrameSize);
                code.decFrameSize(deltaFrameSize);
            }
        }

        public void caseNew(New tree) {
            int size = tree.args.length;
            emitLoadConstant(targetReg, size * WORD_SIZE);
            code.emit(SYSCALL, targetReg, targetReg, SYS_GC_ALLOC);
            int tmpReg = code.getRegister();
            for (int i = 0; i < size; i++) {
                genLoad(tree.args[i], tmpReg);
                code.emit(STW, tmpReg, targetReg, i * WORD_SIZE);
            }
            code.freeRegister(tmpReg);
        }

        public void caseSelect(Select tree) {
            genLoad(tree.prefix, targetReg);
            Code.Label l = code.getLabel();
            code.emit(BNE, targetReg, l);
            emitLoadConstant(1, 1);
            code.emit(SYSCALL, 1, ZERO, SYS_EXIT);
            code.anchorLabel(l);
            code.emit(LDW, targetReg, targetReg, tree.sym.offset);
        }

        public void caseUnop(Unop tree) {
            if (tree.op == Tree.NEG) {
                genLoad(tree.expr, targetReg);
                code.emit(SUB, targetReg, ZERO, targetReg);
            }
            else
                caseDefault(tree);
        }

        public void caseDefault(Tree tree) {
            genLoad(new If(tree.pos,
                           tree,
                           new IntLit(tree.pos, 1),
                           new IntLit(tree.pos, 0)),
                    targetReg);
        }

        private void emitLoadConstant(int reg, int value) {
            int highBits = value >> 16;
            int lowBits = value & 0xFFFF;
            if (highBits != 0) {
                code.emit(ORIU, reg, ZERO, highBits);
                code.emit(LSHI, reg, reg, 16);
                code.emit(ORIU, reg, reg, lowBits);
            } else
                code.emit(ORIU, reg, ZERO, lowBits);
        }
        private int opcode(int op) {
            switch (op) {
            case Tree.ADD: return ADD;
            case Tree.SUB: return SUB;
            case Tree.MUL: return MUL;
            case Tree.DIV: return DIV;
            case Tree.MOD: return MOD;
            default: throw new Error("unknown opcode: " + op);
            }
        }
    }

    /**
     * Visitor generating code for conditions, that is constructs
     * which jump somewhere on the result of a test.
     */
    private class GenCond extends DefaultVisitor implements Visitor, RISC {
        private Code.Label targetLabel;
        private boolean when;

        public void generate(Tree tree, Code.Label targetLabel, boolean when) {
            Code.Label bkpTargetLabel = this.targetLabel;
            this.targetLabel = targetLabel;
            boolean bkpWhen = this.when;
            this.when = when;
            tree.apply(this);
            this.targetLabel = bkpTargetLabel;
            this.when = bkpWhen;
        }

        public void caseBinop(Binop tree) {
            switch (tree.op) {
            case Tree.EQ: case Tree.NE: case Tree.LT:
            case Tree.LE: case Tree.GT: case Tree.GE:
                int leftReg = code.getRegister();
                genLoad(tree.left, leftReg);
                int rightReg = code.getRegister();
                genLoad(tree.right, rightReg);
                
                if (tree.inversed)
                	code.emit(CMP, leftReg, rightReg, leftReg);
                else
                	code.emit(CMP, leftReg, leftReg, rightReg);
                
                code.freeRegister(rightReg);
                code.emit(opcode(tree.op, when), leftReg, targetLabel);                
                code.freeRegister(leftReg);
                break;
            case Tree.OR:
                assert !tree.inversed : "OR Binop has inversed trees!";
                if (when) {
                    genCond(tree.left, targetLabel, true);
                    genCond(tree.right, targetLabel, true);
                } else {
                    Code.Label afterLabel = code.getLabel();
                    genCond(tree.left, afterLabel, true);
                    genCond(tree.right, targetLabel, false);
                    code.anchorLabel(afterLabel);
                }
                break;
            default:
                caseDefault(tree);
            }
        }

        public void caseUnop(Unop tree) {
            if (tree.op == Tree.NOT)
                genCond(tree.expr, targetLabel, !when);
            else
                caseDefault(tree);
        }

        public void caseDefault(Tree tree) {
            int tmpReg = code.getRegister();
            genLoad(tree, tmpReg);
            code.emit((when ? BNE : BEQ), tmpReg, targetLabel);
            code.freeRegister(tmpReg);
        }

        private int opcode(int op, boolean negated) {
            switch (op) {
            case Tree.EQ : return (negated ? BEQ : BNE);
            case Tree.NE : return (negated ? BNE : BEQ);
            case Tree.LT : return (negated ? BLT : BGE);
            case Tree.LE : return (negated ? BLE : BGT);
            case Tree.GT : return (negated ? BGT : BLE);
            case Tree.GE : return (negated ? BGE : BLT);
            default      : throw new Error("invalid operator: " + op);
            }
        }
    }
}
