/*  einsc -- a compiler for eins
 *    Parser
 *
 *    Matthias Zenger, Michel Schinz, Philippe Altherr
 *    $Id: Parser.java,v 1.1 2005/03/15 14:46:53 dragos Exp $
 */

package einsc;

import java.io.InputStreamReader;
import einsc.Tree.*;
import java.util.Vector;

/** This class implements the parser of the compiler.
 */
class Parser extends Scanner {

    public Parser(InputStreamReader in) {
        super(in);
    }

    /** Convert a vector to a tree array.
     */
    private Tree[] toArray(Vector<Tree> trees) {
        return trees.toArray(new Tree[trees.size()]);
    }

    /** Reports an error that a specific token was expected but
     *  a different token found.
     */
    private void error(int expected) {
        error("token of class " + tokenClass(expected));
    }

    /** Reports an error that a specific token was expected but
     *  a different token found.
     */
    private void error(String expected) {
        String message = "syntax error\n" +
            "expected: " + expected + "\n" +
            "found   : " + representation();
        Report.fail(start, message);
        throw new Error();
    }

    /** If the current token corresponds to the token class 'expected'
     *  then skip it, otherwise signal an error.
     */
    private void accept(int expected) {
        if (token == expected)
            nextToken();
        else
            error(expected);
    }

    /** Parse an identifier.
     */
    private String ident() {
        if (token == IDENT) {
            String name = chars;
            nextToken();
            return name;
        } else {
            nextToken();
            return "<error>";
        }
    }

    /** Parse a natural number (positive).
     */
    private int number() {
        if (token == NUMBER) {
            String num = chars;
            nextToken();
            try {
                return Integer.parseInt(num);
            } catch (NumberFormatException e) {
                error("natural number");
                return 0;
            }
        } else {
            nextToken();
            return 0;
        }
    }

    /** The unit literal: "new Unit()"
     */
    private Tree unitLit(int pos) {
        return new New(pos, "Unit", new Tree[0]);
    }

    /** Parse the input and return an abstract syntax tree.
     */
    public Tree parse() {
        Tree result = program();
        accept(EOF);
        return result;
    }

    /** Program = { Declaration } Expression
     */
    private Tree program() {
        int pos = start;
        Vector<Tree> decls = new Vector<Tree>();
        while (token == CLASS || token == DEF) {
            decls.add(declaration());
        }
        return new Program(pos, toArray(decls), expression());
    }

    /** Declaration = FunDecl | ClassDecl
     */
    private Tree declaration() {
        if (token == DEF)
            return funDecl();
        else
            return classDecl();
    }

    /** FunDecl = "def" ident "(" Formals ")" ":" Type "=" Expression ";"
     */
    private Tree funDecl() {
        int pos = start;
        accept(DEF);
        String name = ident();
        accept(LPAREN);
        Tree[] params = formals();
        accept(RPAREN);
        accept(COLON);
        Tree restpe = type();
        accept(EQUAL);
        Tree expr = expression();
        accept(SEMICOLON);
        return new FunDef(pos, name, params, restpe, expr);
    }

    /** Formal = ident ":" Type
     */
    private Formal formal() {
        int pos = start;
        String name = ident();
        accept(COLON);
        return new Formal(pos, name, type());
    }

    /** Formals = [ Formal { "," Formal } ]
     */
    private Tree[] formals() {
        Vector<Tree> args = new Vector<Tree>();
        if (token == IDENT)
            while (true) {
                args.add(formal());
                if (token == COMMA)
                    nextToken();
                else
                    break;
            }
        return toArray(args);
    }

    /** ClassDecl = "class" ident "{" { Field } "}" 
     */
    private Tree classDecl() {
        int pos = start;
        Vector<Tree> fields = new Vector<Tree>();
        accept(CLASS);
        String name = ident();
        accept(LBRACE);
        while (token == VAL)
            fields.add(field());
        accept(RBRACE);
        return new ClassDef(pos, name, toArray(fields));
    }

    /** Field = "val" Formal ";"
     */
    private Tree field() {
        int pos = start;
        accept(VAL);
        Formal field = formal();
        accept(SEMICOLON);
        return field;
    }

    /** Type = "Int" | ident | "(" Types ")" "=>" Type
     */
    private Tree type() {
        int pos = start;
        if (token == INT) {
            nextToken();
            return new IntType(pos);
        }
        else if (token == IDENT) {
            String name = chars;
            nextToken();
            return new TypeIdent(pos,  name);
        } else {
            accept(LPAREN);
            Tree[] argtpes = types();
            accept(RPAREN);
            accept(ARROW);
            return new FunType(pos, argtpes, type());
        }
    }

    /** Types = [ Type { "," Type } ]
     */
    private Tree[] types() {
        Vector<Tree> tpes = new Vector<Tree>();
        if ((token == INT) ||
            (token == IDENT) ||
            (token == LPAREN))
            while (true) {
                tpes.add(type());
                if (token == COMMA)
                    nextToken();
                else
                    break;
            }
        return toArray(tpes);
    }

    /** Statement = "while" "(" Expression ")" "{" { Statement } "}"
     *            | "var" Formal "=" Expression ";"
     *            | "let" ident "=" Expression ";"
     *            | "do" Expression ";"
    */
    private Tree statement() {
        int pos = start;
        if (token == WHILE) {
            nextToken();
            accept(LPAREN);
            Tree cond = expression();
            accept(RPAREN);
            accept(LBRACE);
            Tree[] stats = statements();
            accept(RBRACE);
            return new While(pos, cond, stats);            
        }
        else if (token == VAR) {
            nextToken();
            Formal formal = formal();
            accept(EQUAL);
            Tree init = expression();
            accept(SEMICOLON);
            return new VarDef(pos, formal, init);
        }
        else if (token == LET) {
            nextToken();
            String name = ident();
            accept(EQUAL);
            Tree value = expression();
            accept(SEMICOLON);
            return new Assign(pos, name, value);
        }
        else {
            accept(DO);
            Tree expr = expression();
            accept(SEMICOLON);
            return new Do(pos, expr);
        }
    }
    /** Statements = { Statement }
     */
    private Tree[] statements() {
        Vector<Tree> stats = new Vector<Tree>();
        while (token == WHILE || token == VAR ||
               token == LET || token == DO)
            stats.add(statement());
        return toArray(stats);
    }

    /** Expression = "if" "(" Expression ")" Expression [ "else" Expression ]
     *             | OrExpression
     */
    private Tree expression() {
        int pos = start;
        if (token == IF) {
            nextToken();
            accept(LPAREN);
            Tree cond = expression();
            accept(RPAREN);
            Tree thenpart = expression();
            if (token == ELSE) {
                nextToken();
                return new If(pos, cond, thenpart, expression());
            } else
                return new If(pos, cond, thenpart, unitLit(pos));
        } else
            return orExpression();
    }

    /** OrExpression = AndExpression
     *               | OrExpression "|" AndExpression
     */
    private Tree orExpression() {
        Tree expr = andExpression();
        while (token == OR) {
            int pos = start;
            nextToken();
            expr = new Binop(pos, Tree.OR, expr, andExpression(), false);
        }
        return expr;
    }

    /** Build a logical not.
     */
    private Tree not(Tree tree) {
        return new Unop(tree.pos, Tree.NOT, tree);
    }

    /** AndExpression = CmpExpression
     *                | AndExpression "&" CmpExpression
     */
    private Tree andExpression() {
        Tree expr = cmpExpression();
        while (token == AND) {
            int pos = start;
            nextToken();
            expr = not(new Binop(pos, Tree.OR, not(expr), not(cmpExpression()), false));
        }
        return expr;
    }

    /** CmpExpression = [ SumExpression CompOp ] SumExpression
     *  CompOp        = "=="
     *                | "!="
     *                | "<"
     *                | ">"
     *                | "<="
     *                | ">="
     */
    private Tree cmpExpression() {
        Tree expr = sumExpression();
        int pos = start;
        switch (token) {
            case EQ:
                nextToken();
                return new Binop(pos, Tree.EQ, expr, sumExpression(), false);
            case NE:
                nextToken();
                return new Binop(pos, Tree.NE, expr, sumExpression(), false);
            case LT:
                nextToken();
                return new Binop(pos, Tree.LT, expr, sumExpression(), false);
            case LE:
                nextToken();
                return new Binop(pos, Tree.LE, expr, sumExpression(), false);
            case GT:
                nextToken();
                return new Binop(pos, Tree.GT, expr, sumExpression(), false);
            case GE:
                nextToken();
                return new Binop(pos, Tree.GE, expr, sumExpression(), false);
            default:
                return expr;
        }
    }

    /** SumExpression = Term | SumExpression SumOp Term
     *  SumOp         = "+" | "-"
     */
    private Tree sumExpression() {
        Tree term = term();
        while (true) {
            int pos = start;
            switch (token) {
                case ADD:
                    nextToken();
                    term = new Binop(pos, Tree.ADD, term, term(), false);
                    break;
                case SUB:
                    nextToken();
                    term = new Binop(pos, Tree.SUB, term, term(), false);
                    break;
                default:
                    return term;
            }
        }
    }

    /** Term           = [ Term ProdOp ] [ NegateOp ] Factor
     *  NegateOp       = "-" | "!"
     *  ProdOp         = "*" | "/" | "%"
     */
    private Tree term() {
        Tree term = signedFactor();
        while (true) {
            int pos = start;
            switch (token) {
                case MUL:
                    nextToken();
                    term = new Binop(pos, Tree.MUL, term, signedFactor(), false);
                    break;
                case DIV:
                    nextToken();
                    term = new Binop(pos, Tree.DIV, term, signedFactor(), false);
                    break;
                case MOD:
                    nextToken();
                    term = new Binop(pos, Tree.MOD, term, signedFactor(), false);
                    break;
                default:
                    return term;
            }
        }
    }

    private Tree signedFactor() {
        int pos = start;
        if (token == SUB) {
            nextToken();
            return new Unop(pos, Tree.NEG, appliedFactor());
        } else if (token == NOT) {
            nextToken();
            return new Unop(pos, Tree.NOT, appliedFactor());
        } else
            return appliedFactor();
    }

    private Tree appliedFactor() {
        Tree factor = factor();
        while (token == LPAREN || token == DOT)
            if (token == LPAREN)
                factor = new Apply(start, factor, params());
            else {
                accept(DOT);
                String name = chars;
                accept(IDENT);
                factor = new Select(start, factor, name);
            }
        return factor;
    }

    /** Factor = ident
     *         | number
     *         | string
     *         | "true"
     *         | "false"
     *         | "null"
     *         | "(" Expression ")"
     *         | "{" { Statement } [ Expression ] "}"
     *         | Factor Params
     *         | "new" ident Params
     *         | Factor "." ident
     */
    private Tree factor() {
        int pos = start;
        switch (token) {
        case IDENT:
            return new Ident(pos, ident());
        case NUMBER:
            return new IntLit(pos, number());
        case STRING:
            String str = chars;
            Tree expr = new NullLit(pos);
            for (int i = str.length() - 1; i >= 0; i--)
                expr = new New(pos, "List",
                               new Tree[] { new IntLit(pos, str.charAt(i)),
                                            expr });
            nextToken();
            return expr;
        case TRUE:
            nextToken();
            return new IntLit(pos, 1);
        case FALSE:
            nextToken();
            return new IntLit(pos, 0);
        case NULL:
            accept(NULL);
            return new NullLit(pos);            
        case LPAREN:
            accept(LPAREN);
            Tree nested = expression();
            accept(RPAREN);
            return nested;
        case LBRACE:
            accept(LBRACE);
            Tree[] stats = statements();
            Tree main =
                token != RBRACE ? expression() : unitLit(pos);            
            accept(RBRACE);
            return
                stats.length > 0 ?
                new Block(pos, stats, main) :
                main;
        case NEW:
            accept(NEW);
            String name = ident();
            return new New(pos, name, params());
        default:
            error("factor");
            nextToken();
            return new Bad(pos);
        }
    }

    /** Params      = "(" Expressions ")"
     *  Expressions = [ Expression { "," Expression } ]
     */
    private Tree[] params() {
        accept(LPAREN);
        Vector<Tree> exprs = new Vector<Tree>();
        if (token != RPAREN)
            while (true) {
                exprs.add(expression());
                if (token == COMMA)
                    nextToken();
                else
                    break;
            }
        accept(RPAREN);
        return toArray(exprs);
    }
}
