/*  einsc -- a compiler for eins
 *    Scanner
 *
 *    Matthias Zenger, Michel Schinz, Philippe Altherr
 *    $Id: Scanner.java,v 1.1 2005/03/15 14:46:51 dragos Exp $
 */

package einsc;

import java.io.InputStreamReader;
import java.io.IOException;

/** This class implements the scanner of the compiler.
 */
class Scanner implements Tokens {

    /** This character represents the end of the input.
     */
    private static final char EOF_CH = (char)-1;

    /** The current token class.
     */
    protected int token;

    /** The position of the first character of the current token.
     */
    protected int start;

    /** The string representation of the current token. This variable
     *  is only set if the current token class has several possible
     *  textual representations.
     */
    protected String chars;

    /** A buffer for constructing string representations of tokens.
     */
    private StringBuffer buf;

    /** The current character.
     */
    private char ch;

    /** The line and the column of the current token.
     */
    private int line = 1;
    private int column = 0;

    /** The input stream providing the input characters.
     */
    private final InputStreamReader in;

    Scanner(InputStreamReader in) {
        this.in = in;
        this.buf = new StringBuffer();
        nextCh();
        nextToken();
    }

    /** close the file input stream
     */
    public void close() throws IOException {
        in.close();
    }

    /** This method reads the next token and stores the token class
     *  in variable 'token'. If the token representation is not unique
     *  it will also leave a textual representation in variable 'chars'.
     *  Whitespaces and comments are skipped by this method.
     */
    public void nextToken() {
        // skip comments and whitespace characters

        while (true) {
            while (Character.isWhitespace(ch))
                nextCh();
            if (ch == '/') {
                start = Position.encode(line, column);
                nextCh();
                if (ch == '/') {
                    while (ch != '\n' && ch != EOF_CH)
                        nextCh();
                } else {
                    token = DIV;
                    return;
                }
            } else
                break;
        }

        // initialize the position of the current token
        start = Position.encode(line, column);
        // read the current token
        token = readToken();
    }

    /** Read the next token, store its representation (if its not
     *  unique) in variable 'chars' and return the token class.
     */
    private int readToken() {

        switch (ch) {
            // end of file
            case EOF_CH:
                return EOF;
            // identifier or keyword
            case 'A': case 'B': case 'C': case 'D': case 'E': case 'F':
            case 'G': case 'H': case 'I': case 'J': case 'K': case 'L':
            case 'M': case 'N': case 'O': case 'P': case 'Q': case 'R':
            case 'S': case 'T': case 'U': case 'V': case 'W': case 'X':
            case 'Y': case 'Z':
            case 'a': case 'b': case 'c': case 'd': case 'e': case 'f':
            case 'g': case 'h': case 'i': case 'j': case 'k': case 'l':
            case 'm': case 'n': case 'o': case 'p': case 'q': case 'r':
            case 's': case 't': case 'u': case 'v': case 'w': case 'x':
            case 'y': case 'z':
                buf.setLength(0);
                do {
                    buf.append(ch);
                    nextCh();
                } while (Character.isLetterOrDigit(ch) || ch == '_');
                chars = buf.toString();
                
                if      (chars.equals("class")) return CLASS;
                else if (chars.equals("val"))   return VAL;
                else if (chars.equals("new"))   return NEW;
                else if (chars.equals("null"))  return NULL;
                else if (chars.equals("let"))   return LET;
                else if (chars.equals("do"))    return DO;
                else if (chars.equals("def"))   return DEF;
                else if (chars.equals("var"))   return VAR;
                else if (chars.equals("while")) return WHILE;
                else if (chars.equals("if"))    return IF;
                else if (chars.equals("else"))  return ELSE;
                else if (chars.equals("true"))  return TRUE;
                else if (chars.equals("false")) return FALSE;
                else if (chars.equals("Int"))   return INT;
                else
                    return IDENT;
            // zero
            case '0':
                chars = "0";
                nextCh();
                return NUMBER;
            // non-zero numbers
            case '1': case '2': case '3': case '4':
            case '5': case '6': case '7': case '8': case '9':
                buf.setLength(0);
                do {
                    buf.append(ch);
                    nextCh();
                } while (Character.isDigit(ch));
                chars = buf.toString();
                return NUMBER;
            // strings
            case '"':
                nextCh();
                buf.setLength(0);
                while (ch != '"') {
                    if (ch == '\n' || ch == EOF_CH)
                        Report.fail(start, "string not properly terminated");
                    buf.append(ch);
                    nextCh();
                }
                nextCh();
                chars = buf.toString();
                return STRING;
            // operators and special characters
            case '.':
                nextCh();
                return DOT;
            case '(':
                nextCh();
                return LPAREN;
            case ')':
                nextCh();
                return RPAREN;
            case '{':
                nextCh();
                return LBRACE;
            case '}':
                nextCh();
                return RBRACE;
            case ',':
                nextCh();
                return COMMA;
            case ';':
                nextCh();
                return SEMICOLON;
            case ':':
                nextCh();
                return COLON;
            case '=':
                nextCh();
                if (ch == '=') {
                    nextCh();
                    return EQ;
                } else if (ch == '>') {
                    nextCh();
                    return ARROW;
                } else
                    return EQUAL;
            case '!':
                nextCh();
                if (ch == '=') {
                    nextCh();
                    return NE;
                } else
                    return NOT;
            case '<':
                nextCh();
                if (ch == '=') {
                    nextCh();
                    return LE;
                } else
                    return LT;
            case '>':
                nextCh();
                if (ch == '=') {
                    nextCh();
                    return GE;
                } else
                    return GT;
            case '+':
                nextCh();
                return ADD;
            case '-':
                nextCh();
                return SUB;
            case '*':
                nextCh();
                return MUL;
            case '/':
                nextCh();
                return DIV;
            case '%':
                nextCh();
                return MOD;
            case '&':
                nextCh();
                return AND;
            case '|':
                nextCh();
                return OR;
            default:
                Report.fail(start, "invalid character: '" + ch + "'");
                return BAD;
        }
    }

    /** Return a textual representation of the current token.
     */
    public String representation() {
        String representation = tokenClass(token);
        if (token == NUMBER || token == IDENT || token == STRING)
            representation += "(" + chars + ")";
        return representation;
    }

    /** Return the textual representation of the given token class.
     */
    public static String tokenClass(int token) {
        switch (token) {
            case EOF:
                return "<eof>";
            case BAD:
                return "<bad>";
            case IDENT:
                return "ident";
            case NUMBER:
                return "number";
            case STRING:
                return "string";
            case LPAREN:
                return "(";
            case RPAREN:
                return ")";

            case LBRACE:
                return "{";
            case RBRACE:
                return "}";
            case COMMA:
                return ",";
            case SEMICOLON:
                return ";";
            case COLON:
                return ":";
            case EQUAL:
                return "=";
            case EQ:
                return "==";
            case NOT:
                return "!";
            case NE:
                return "!=";
            case LT:
                return "<";
            case LE:
                return "<=";
            case GT:
                return ">";
            case GE:
                return ">=";
            case ARROW:
                return "=>";
            case ADD:
                return "+";
            case SUB:
                return "-";
            case MUL:
                return "*";
            case DIV:
                return "/";
            case MOD:
                return "%";
            case AND:
                return "&";
            case OR:
                return "|";
             case DOT:
                return ".";
            case DEF:
                return "def";
            case VAR:
                return "var";
            case WHILE:
                return "while";
            case IF:
                return "if";
            case ELSE:
                return "else";
            case TRUE:
                return "true";
            case FALSE:
                return "false";
            case CLASS:
                return "class";
            case VAL:
                return "val";
            case NEW:
                return "new";
            case NULL:
                return "null";
            case INT:
                return "Int";
            case LET:
                return "let";
            case DO:
                return "do";
            default:
                throw new Error("unknown token class " + token);
        }
    }

    /** Puts the next character into 'ch' and updates the current position.
     */
    private void nextCh() {
        switch (ch) {
            case EOF_CH:
                return;
            case '\n':
                column = 1;
                line++;
                break;
            default:
                column++;
        }
        try {
            ch = (char)in.read();
            oldch = ((oldch == '\r') && (ch == '\n')) ? (char)in.read() : ch;
            ch = (oldch == '\r') ? '\n' : oldch;
        } catch (IOException e) {
            Report.fail(Position.encode(line, column), e.getMessage());
        }
    }

    private char oldch;
}
