/*  einsc -- a compiler for eins
 *    Representation of the abstract syntax
 *
 *    Matthias Zenger, Michel Schinz, Philippe Altherr
 *    $Id: Tree.java,v 1.1 2005/03/15 14:46:49 dragos Exp $
 */

package einsc;


/** Class Tree defines the type of all abstract syntax tree nodes.
 */
abstract public class Tree {

    /** The source code position of this tree node.
     */
    public int pos;

    /** The type of the tree.
     */
    public Type type;
    
    /** The number of registers needed to compile this subtree. 
     */
    public int regNeed;

    /** Constructor for tree nodes.
     */
    Tree(int pos) {
        this.pos = pos;
    }

    /** Applies the given visitor object to this tree node.
     */
   public  abstract void apply(Visitor v);
    
    /** This node represents an erroneous syntactical construct.
     */
    public static class Bad extends Tree {
        public Bad(int pos) {
            super(pos);
        }
        
        public void apply(Visitor v) {
            v.caseBad(this);
        }
    }

    /** P = Program { D } E
     */
    public static class Program extends Tree {
        public Tree[] decls;
        public Tree main;
        
        public Program(int pos, Tree[] decls, Tree main) {
            super(pos);
            this.decls = decls;
            this.main = main;
        }

        public void apply(Visitor v) {
            v.caseProgram(this);
        }
    }
    
    /** D = ClassDef ident { F }
     */
    public static class ClassDef extends Tree {
        public String name;
        public Tree[] fields;
        public Symbol sym;

        public ClassDef(int pos, String name, Tree[] fields) {
            super(pos);
            this.name = name;
            this.fields = fields;
        }

        public void apply(Visitor v) {
            v.caseClassDef(this);
        }
    }

    /** D = FunDef ident { F } T E
     */
    public static class FunDef extends Tree {
        public String name;
        public Tree[] params;
        public Tree restype;
        public Tree body;
        public Symbol sym;

        public FunDef(int pos, String name, Tree[] params, Tree restype, Tree body) {
            super(pos);
            this.name = name;
            this.params = params;
            this.restype = restype;
            this.body = body;
        }

        public void apply(Visitor v) {
            v.caseFunDef(this);
        }
    }


    /** F = Formal ident T
     */
    public static class Formal extends Tree {
        public String name;
        public Tree vartype;
        public Symbol sym;

        public Formal(int pos, String name, Tree vartype) {
            super(pos);
            this.name = name;
            this.vartype = vartype;
        }

        public void apply(Visitor v) {
            v.caseFormal(this);
        }
    }

    /** T = IntType
     */
    public static class IntType extends Tree {
        
        public IntType(int pos) {
            super(pos);
        }

        public void apply(Visitor v) {
            v.caseIntType(this);
        }
    }

    /** T = TypeIdent ident
     */
    public static class TypeIdent extends Tree {
        public String name;
        public Symbol sym;

        public TypeIdent(int pos, String name) {
            super(pos);
            this.name = name;
        }
        
        public void apply(Visitor v) {
            v.caseTypeIdent(this);
        }
    }

    /** T = FunType { T } T
     */
    public static class FunType extends Tree {
        public Tree[] paramtypes;
        public Tree restype;

        public FunType(int pos, Tree[] paramtypes, Tree restype) {
            super(pos);
            this.paramtypes = paramtypes;
            this.restype = restype;
        }

        public void apply(Visitor v) {
            v.caseFunType(this);
        }
    }

    /** S = While E { S }
     */
    public static class While extends Tree {
        public Tree cond;
        public Tree[] stats;

        public While(int pos, Tree cond, Tree[] stats) {
            super(pos);
            this.cond = cond;
            this.stats = stats;
        }

        public void apply(Visitor v) {
            v.caseWhile(this);
        }
    }

    /** S = VarDef F E
     */
    public static class VarDef extends Tree {
        public Formal variable;
        public Tree init;
        
        public VarDef(int pos, Formal variable, Tree init) {
            super(pos);
            this.variable = variable;
            this.init = init;
        }
        
        public void apply(Visitor v) {
            v.caseVarDef(this);
        }
    }

    /** S = Assign ident E
     */
    public static class Assign extends Tree {
        public String name;
        public Tree expr;
        public Symbol sym;

        public Assign(int pos, String name, Tree expr) {
            super(pos);
            this.name = name;
            this.expr = expr;
        }

        public void apply(Visitor v) {
            v.caseAssign(this);
        }
    }

    /** S = Do E
     */
    public static class Do extends Tree {
        public Tree expr;

        public Do(int pos, Tree expr) {
            super(pos);
            this.expr = expr;
        }

        public void apply(Visitor v) {
            v.caseDo(this);
        }
    }

    /** E = If E E E
     */
    public static class If extends Tree {
        public Tree cond;
        public Tree thenp;
        public Tree elsep;

        public If(int pos, Tree cond, Tree thenp, Tree elsep) {
            super(pos);
            this.cond = cond;
            this.thenp = thenp;
            this.elsep = elsep;
        }

        public void apply(Visitor v) {
            v.caseIf(this);
        }
    }

    
    /** E = Binop B E E
     */
    public static class Binop extends Tree {
        public int op;
        public Tree left;
        public Tree right;
        public boolean inversed;

        public Binop(int pos, int op, Tree left, Tree right, boolean inversed) {
            super(pos);
            this.op = op;
            this.left = left;
            this.right = right;
            this.inversed = inversed;
        }

        public void apply(Visitor v) {
            v.caseBinop(this);
        }
    }
    
    /** B = Add | Sub | Mul | Div | Mod
     *    | Eq | Ne | Lt | Le | Gt | Ge | Or
     */
    public static final int ADD =  0;
    public static final int SUB =  1;
    public static final int MUL =  2;
    public static final int DIV =  3;
    public static final int MOD =  4;
    public static final int EQ  =  5;
    public static final int NE  =  6;
    public static final int LT  =  7;
    public static final int LE  =  8;
    public static final int GT  =  9;
    public static final int GE  = 10;
    public static final int OR  = 11;

    /** E = Unop U E
     */
    public static class Unop extends Tree {
        public int op;
        public Tree expr;
        
        public Unop(int pos, int op, Tree expr) {
            super(pos);
            this.op = op;
            this.expr = expr;
        }
        
        public void apply(Visitor v) {
            v.caseUnop(this);
        }
    }

    /** U = Not | Neg
     */
    public static final int NOT =  12;
    public static final int NEG =  13;

    /** E = Ident ident
     */
    public static class Ident extends Tree {
        public String name;
        public Symbol sym;

        public Ident(int pos, String name) {
            super(pos);
            this.name = name;
        }

        public void apply(Visitor v) {
            v.caseIdent(this);
        }
    }

    /** E = NullLit
     */
    public static class NullLit extends Tree {
        
        public NullLit(int pos) {
            super(pos);
        }
        
        public void apply(Visitor v) {
            v.caseNullLit(this);
        }
    }

    /** E = IntLit
     */
    public static class IntLit extends Tree {
        public int value;
        
        public IntLit(int pos, int value) {
            super(pos);
            this.value = value;
        }

        public void apply(Visitor v) {
            v.caseIntLit(this);
        }
    }
    
    /** E = Apply E { E }
     */
    public static class Apply extends Tree {
        public Tree fun;
        public Tree[] args;
        
        public Apply(int pos, Tree fun, Tree[] args) {
            super(pos);
            this.fun = fun;
            this.args = args;
        }

        public void apply(Visitor v) {
            v.caseApply(this);
        }
    }

    /** E = Block { S } E
     */
    public static class Block extends Tree {
        public Tree[] stats;
        public Tree main;
        
        public Block(int pos, Tree[] stats, Tree main) {
            super(pos);
            this.stats = stats;
            this.main = main;
        }

        public void apply(Visitor v) {
            v.caseBlock(this);
        }
    }

    /** E = New ident { E }
     */
    public static class New extends Tree {
        public String name;
        public Tree[] args;
        public Symbol sym;

        public New(int pos, String name, Tree[] args) {
            super(pos);
            this.name = name;
            this.args = args;
        }

        public void apply(Visitor v) {
            v.caseNew(this);
        }
    }

    /** E = Select E ident 
     */
    public static class Select extends Tree {
        public Tree prefix;
        public String selector;
        public Symbol sym;

        public Select(int pos, Tree prefix, String selector) {
            super(pos);
            this.prefix = prefix;
            this.selector = selector;
        }

        public void apply(Visitor v) {
            v.caseSelect(this);
        }
    }

}
